package algo.branch;

import algo.datastructure.*;

import java.util.*;

public class Packing {

    static PartSetVertices unpackedVertices;
    static int t;
    static Graph g;
    public static Set<Pack> packs = new HashSet<>();

    static boolean updateEverything = true;
    public static boolean useWeightedCost;

    static void initializePacking(Graph g, int t, boolean useWeightedCost) {
        unpackedVertices = new PartSetVertices(g);
        Packing.g = g;
        Packing.t = t;
        Packing.useWeightedCost = useWeightedCost;
        for (Vertex vertex : g) {
            vertex.inManyPacks = 0;
            unpackedVertices.add(vertex);
        }
        packs.clear();
    }
    static public int improvePackingBySwaps() {
        int increase = 0;
        List<Pack> packsList = new ArrayList<>(packs);
        //swap between packs
        for (int i = 0; i < packsList.size(); i++) {
            Pack pack1 = packsList.get(i);
            Edge minEdge1 = findMinEdge(pack1);
            if (minEdge1 == null) {
                LowerBound.flagUnfeasible = true;
                return 1;
            }
            for (int j = i + 1; j < packsList.size(); j++) {
                Pack pack2 = packsList.get(j);
                Edge minEdge2 = findMinEdge(pack2);
                if (minEdge2 == null) {
                    LowerBound.flagUnfeasible = true;
                    return 1;
                }
                if (swapIfImprovePacking(minEdge1, minEdge2, pack1, pack2)) {
                    pack1.computeCost(g);
                    pack2.computeCost(g);
                    break;
                }
            }
        }
        //swap between pack and unpacked vertices
        for (Pack pack : packsList) {
            Edge minEdge = findMinEdge(pack);
            if (minEdge == null) {
                LowerBound.flagUnfeasible = true;
                return 1;
            }
            Vertex u = minEdge.u;
            Vertex v = minEdge.v;

            for (Vertex w : unpackedVertices) {
                int costOfPack = pack.cost;
                //if w and v in same NC then w and u may not be swapped and vice versa
                if (w == u || w == v) continue;
                int costToU = (g.inSameNC(w, v)) ? -1 : g.getEdge(w, u).getResolveCostOfNonCommons();
                int costToV = (g.inSameNC(w, u)) ? -1 : g.getEdge(w, v).getResolveCostOfNonCommons();

                if (costToU <= costOfPack && costToV <= costOfPack) continue;//no improvement can be done with w

                int minCostOfWToPack = Math.max(costToU, costToV);//the minimum to other vertices of pack, and the maximum to u and v since the swap with u or v can be chosen
                Vertex[] packedVertices = pack.packedVertices;
                boolean wIsInPack = false;

                for (int j = 0; j < pack.size; j++) {
                    Vertex packedVertex = packedVertices[j];
                    if (packedVertex == u || packedVertex == v) continue;
                    if (packedVertex == w) {
                        wIsInPack = true;
                        break;
                    }
                    minCostOfWToPack = Math.min(minCostOfWToPack, g.getEdge(packedVertex, w).getResolveCostOfNonCommons());
                    if (g.inSameNC(packedVertex, w)) {
                        minCostOfWToPack = -1;
                        break;
                    }
                    if (minCostOfWToPack <= costOfPack) break;
                }
                if (wIsInPack) continue;//w is in one neighborhood class with a other vertex of pack

                if (minCostOfWToPack <= costOfPack) continue;//no cost increase

                costToU = Math.min(minCostOfWToPack, costToU);
                costToV = Math.min(minCostOfWToPack, costToV);
                if (costToU > costToV) {//swap with v
                    pack.removeVertex(v);
                    pack.addVertex(w);
                    if (!unpackedVertices.contains(v)) unpackedVertices.add(v);
                    if (w.inManyPacks == w.weight) unpackedVertices.remove(w);
                    pack.computeCost(g);
                    increase += pack.cost - costOfPack;
                    break;
                } else {//swap with u
                    pack.removeVertex(u);
                    pack.addVertex(w);
                    if (!unpackedVertices.contains(u)) unpackedVertices.add(u);
                    if (w.inManyPacks == w.weight) unpackedVertices.remove(w);
                    pack.computeCost(g);
                    increase += pack.cost - costOfPack;
                    break;
                }
            }
        }
        return increase;
    }

    private static boolean swapIfImprovePacking(Edge e1, Edge e2, Pack pack1, Pack pack2) {
        if (swapIfImprovePacking(e1.u, e2.u, pack1, pack2))
            return true;
        if (swapIfImprovePacking(e1.u, e2.v, pack1, pack2))
            return true;
        if (swapIfImprovePacking(e1.v, e2.u, pack1, pack2))
            return true;
        if (swapIfImprovePacking(e1.v, e2.v, pack1, pack2))
            return true;
        return false;
    }

    private static boolean swapIfImprovePacking(Vertex u, Vertex v, Pack pack1, Pack pack2) {
        int minCostU = Integer.MAX_VALUE;

        Vertex[] packedVertices = pack2.packedVertices;
        for (int i = 0; i < pack2.size; i++) {
            Vertex p2 = packedVertices[i];
            if (p2 == v) continue;
            if (p2 == u) return false;//u can be element of both packs
            Edge e = g.getEdge(u, p2);
            if (!e.isApart && !e.isNotMergeableWithLabel())
                minCostU = Math.min(minCostU, e.getResolveCostOfNonCommons());
        }

        int minCostV = Integer.MAX_VALUE;
        packedVertices = pack1.packedVertices;
        for (int i = 0; i < pack1.size; i++) {
            Vertex p1 = packedVertices[i];
            if (p1 == u) continue;
            if (p1 == v) return false;//v can be element of both packs
            Edge e = g.getEdge(v, p1);
            if (!e.isApart && !e.isNotMergeableWithLabel())
                minCostV = Math.min(minCostV, e.getResolveCostOfNonCommons());
        }

        if (minCostU > pack2.cost && minCostV > pack1.cost) {
            pack1.removeVertex(u);
            pack2.removeVertex(v);
            pack1.addVertex(v);
            pack2.addVertex(u);
            return true;
        }
        return false;
    }

    private static Edge findMinEdge(Pack pack) {
        //find Vertex pair smallest cost
        int costOfPack = pack.cost;
        Vertex[] packedVertices = pack.packedVertices;
        for (int i = 0; i < pack.size; i++) {
            Vertex u = packedVertices[i];
            for (int j = i + 1; j < pack.size; j++) {
                Vertex v = packedVertices[j];
                Edge e = g.getEdge(u, v);
                int cost = e.getResolveCostOfNonCommons();
                if (e.isApart || e.isNotMergeableWithLabel()) continue;
                if (cost == costOfPack) return e;
            }
        }
        return null;
    }

    static public Pack addPackGreedy() {
        Pack p = new Pack(t);
        Edge startEdge = p.findBestStartEdge(g, unpackedVertices);
        if (startEdge == null) return p;
        unpackedVertices.remove(startEdge.u);
        unpackedVertices.remove(startEdge.v);
        while (p.size != t + 1) {
            boolean added = addOneVertexGreedy(p, g);
            if (!added) {//nothing added -> undo packing
                for (int i = 0; i < p.size; i++) {
                    Vertex packedVertex = p.packedVertices[i];
                    unpackedVertices.add(packedVertex);
                    packedVertex.removeLastPack();
                }
                return p;
            }
        }
        packs.add(p);
        return p;
    }

    static boolean addOneVertexGreedy(Pack p, Graph g) {
        //find best vertex
        Vertex bestVertex = null;
        int bestValue = -1;
        for (Vertex vertex : unpackedVertices) {
            int value = p.minConnectionToPack(g, vertex);
            if (value == -1) continue;
            if (value > bestValue) {
                bestValue = value;
                bestVertex = vertex;
            }
        }
        //nothing found
        if (bestVertex == null) {
            for (int i = 0; i < p.size; i++)
                unpackedVertices.add(p.packedVertices[i]);
            return false;
        }
        //add vertex
        p.addVertex(bestVertex);
        unpackedVertices.remove(bestVertex);
        if (bestValue == -1) bestValue = Integer.MAX_VALUE;
        p.cost = Math.min(p.cost, bestValue);
        return true;
    }

    static public void removePack(Pack pack) {
        Vertex[] packedVertices = pack.packedVertices;
        int start = pack.size - 1;
        for (int i = start; i >= 0; i--) {
            Vertex packedVertex = packedVertices[i];
            pack.removeVertex(packedVertex);
            if (!unpackedVertices.contains(packedVertex))
                unpackedVertices.add(packedVertex);
        }
        packs.remove(pack);
        pack.isDeleted = true;
    }

    static public void updatePackingByAddNonCommon(Edge e) {
        if (LowerBound.reset() || Packing.updateEverything) return;
        List<Pack> commonPacks = inSamePack(e);
        if (commonPacks == null) return;
        for (Pack pack : commonPacks) pack.updatePackCount = true;
    }

    static public void updatePackingByRemoveNonCommon(Edge e) {
        if (LowerBound.reset() || Packing.updateEverything) return;
        List<Pack> commonPacks = inSamePack(e);
        if (commonPacks == null) return;
        for (Pack pack : commonPacks) {
            int newCost = e.getResolveCostOfNonCommons();
            if (e.isApart || e.isNotMergeableWithLabel()
//                    || pack.updatePackCount
            ) continue;
            pack.computeCost(g);
            if (pack.cost > newCost) pack.cost = newCost;
            if (pack.cost == 0) removePack(pack);
        }
    }

    static public void updatePackingByUpdateNotMergeAble(Edge e, boolean notMergeAble) {
        if (LowerBound.reset() || Packing.updateEverything) return;
        List<Pack> commonPacks = inSamePack(e);
        if (commonPacks == null) return;

        if (notMergeAble) {
            for (Pack pack : commonPacks)
                if (e.getResolveCostOfNonCommons() == pack.cost) pack.updatePackCount = true;
        } else {
            //cost can increase since e has now to considered
            for (Pack pack : commonPacks) {
                if (pack.updatePackCount) continue;
                int newCost = e.getResolveCostOfNonCommons();
                if (e.isApart || e.isNotMergeableWithLabel()) return;
                if (newCost < pack.cost) pack.cost = newCost;
                if (pack.cost == 0) removePack(pack);
            }
        }
    }

    static public void updatePackingBySetApart(Edge e) {
        updatePackingByUpdateNotMergeAble(e, e.isApart);
    }

    static public void updatePackingByChangeLabel(Vertex v, Label oldLabel, Label newLabel) {
        if (LowerBound.reset() || Packing.updateEverything) return;
        Pack[] packsOfV = v.packs;
        int start = v.inManyPacks - 1;
        for (int i = start; i >= 0; i--) {
            Pack pack = packsOfV[i];
            if (pack == null) continue;//sometimes more than oe pack can be deleted in updatePackingByUpdateNotMergeAble
            //find mergeable changes and update accordingly
            for (Vertex packedVertex : pack.packedVertices) {
                if (packedVertex == v) continue;
                if (pack.isDeleted) break;
                Edge e = g.getEdge(packedVertex, v);
                boolean oldNotMergeAble = e.isNotMergeableWithLabel();
                v.label = newLabel;
                boolean newNotMergeAble = e.isNotMergeableWithLabel();
                if (newNotMergeAble != oldNotMergeAble)
                    updatePackingByUpdateNotMergeAble(e, newNotMergeAble);
                v.label = oldLabel;
            }
        }
    }

    static public void updatePackingByMerge() {
        updateEverything = true;
    }

    static public void updatePackingByApart() {
        updateEverything = true;
    }

    static private List<Pack> inSamePack(Edge e) {
        //return all packs where vertices of e are element of
        Vertex u = e.u;
        Vertex v = e.v;
        List<Pack> list = new LinkedList<>();
        for (int i = 0; i < u.inManyPacks; i++) {
            Pack p1 = u.packs[i];
            for (int j = 0; j < v.inManyPacks; j++) {
                Pack p2 = v.packs[j];
                if (p2 == p1) list.add(p1);
            }
        }
        if (list.size() == 0)
            return null;
        else return list;
    }
}
