package algo.branch;

import algo.datastructure.*;

import java.util.*;

import static algo.branch.BranchMerge.*;

public class SameNCReductionRule {

    static private boolean abortFlag = false;
    static boolean appliedFlag = false;

    static boolean apply(Vertex u, Vertex v, boolean clique, int cost) {
        Label label = (clique) ? Label.ADJACENT : Label.SEPARATE;
        abortFlag = false;
        appliedFlag = false;

        Set<Vertex> vertices = getAdditionalVerticesToMerge(u, v, label, clique);
        if (abortFlag || vertices.size() == 0) return false;//are not mergeable or no more vertices can be merged

        List<Vertex> verticesToMerge = new ArrayList<>(vertices.size() + 2);
        verticesToMerge.addAll(vertices);
        verticesToMerge.add(u);
        verticesToMerge.add(v);

        // edit edit edges between verticesToMerge and set labels
        int k = budget - cost;
        List<Edge> editedEdges = new ArrayList<>();
        int costForClique = addBetweenEdgesToEditAndReturnCost(g, clique, verticesToMerge, k, editedEdges);
        if (k < costForClique) {
            appliedFlag = true;
            return false;
        }
        boolean[] undoLabel = setLabels(verticesToMerge, clique);
        for (Edge editedEdge : editedEdges) g.editEdge(editedEdge);//need to be unedited afterwards

        //find all non commons to resolve
        vertices.add(u);
        vertices.add(v);
        Vertex[] nonCommonArray = getNonCommonArray(g, vertices, verticesToMerge);
        int minCostResolvingNonCommons = minCostForResolvingNonCommons(g, verticesToMerge, nonCommonArray);
        if (k < minCostResolvingNonCommons + costForClique) abortFlag = true;

        //resolve all non commons
        if (!abortFlag && branchUntilNoNonCommons(verticesToMerge, 0, nonCommonArray, cost + costForClique))
            return true;
        //undo changes
        undoLabels(undoLabel, verticesToMerge);
        for (int i = editedEdges.size() - 1; i >= 0; i--) g.editEdge(editedEdges.get(i));
        appliedFlag = true;
        return false;
    }

    private static boolean[] setLabels(List<Vertex> verticesToMerge, boolean clique) {
        boolean[] undoLabels = new boolean[verticesToMerge.size() - 2];
        for (int i = 0; i < verticesToMerge.size() - 2; i++) {
            Vertex vertex = verticesToMerge.get(i);
            undoLabels[i] = BranchMerge.setLabel(vertex, clique);
        }
        return undoLabels;
    }

    private static void undoLabels(boolean[] undoLabel, List<Vertex> verticesToMerge) {
        for (int i = 0; i < undoLabel.length; i++)
            if (undoLabel[i]) BranchMerge.undoLabel(verticesToMerge.get(i));
    }

    static boolean branchUntilNoNonCommons(List<Vertex> vertices, int index, Vertex[] nonCommons, int cost) {
        BranchMerge.branchTreeNodeCounter++;
        int k = budget - cost;
        if (lowerBound(k) > k) return false;
        if (k < 0) return false;

        if (index == nonCommons.length) {
            Edge[][] revertNonCommons = new Edge[vertices.size() - 1][];
            Vertex u = vertices.get(0);
            for (int i = 1; i < vertices.size(); i++) {//merge all and store the deleted nonCommons
                Vertex vertex = vertices.get(i);
                revertNonCommons[i - 1] = merge(u, vertex);
            }

            if (branch(cost)) return true;

            for (int i = vertices.size() - 1; i >= 1; i--) apart(u, vertices.get(i), revertNonCommons[i - 1]);

            return false;
        }
        Vertex nonCommon = nonCommons[index];//get last nonCommon in array

        Edge[] edgesToNodes = new Edge[vertices.size()];
        for (int i = 0; i < vertices.size(); i++) edgesToNodes[i] = g.getEdge(nonCommon, vertices.get(i));

        //case a: delete all edges
        int costOfCaseA = 0;
        boolean[] editsOfCasesA = new boolean[vertices.size()];
        boolean[] edgeBlocksOfCaseA = new boolean[vertices.size()];

        boolean continueCaseA = true;
        for (Edge e : edgesToNodes) {
            if (e.isBlocked && e.isEdge) {
                continueCaseA = false;
                break;
            }
        }
        if (continueCaseA) {
            for (int i = 0; i < edgesToNodes.length; i++) {
                Edge e = edgesToNodes[i];
                if (e.isEdge) {
                    editsOfCasesA[i] = true;
                    costOfCaseA += e.getEditCost();
                    g.editEdge(e);
                }
                if (!e.isBlocked) {
                    e.block();
                    edgeBlocksOfCaseA[i] = true;
                }
            }

            if (branchUntilNoNonCommons(vertices, index + 1, nonCommons, cost + costOfCaseA)) return true;

            for (int i = editsOfCasesA.length - 1; i >= 0; i--) {
                Edge e = edgesToNodes[i];
                if (edgeBlocksOfCaseA[i]) e.unblock();
                if (editsOfCasesA[i]) {
                    g.editEdge(e);
                }
            }
        }

        //case b:add all edges
        int costOfCaseB = 0;
        boolean[] editsOfCasesB = new boolean[vertices.size()];
        boolean[] blocksOfCasesB = new boolean[vertices.size()];

        boolean continueCaseB = true;
        for (Edge e : edgesToNodes) {
            if (e.isBlocked && !e.isEdge) {
                continueCaseB = false;
                break;
            }
        }

        if (continueCaseB) {
            for (int i = 0; i < edgesToNodes.length; i++) {
                Edge e = edgesToNodes[i];
                if (!e.isEdge) {
                    editsOfCasesB[i] = true;
                    costOfCaseB += e.getEditCost();
                    g.editEdge(e);
                }
                if (!e.isBlocked) {
                    e.block();
                    blocksOfCasesB[i] = true;
                }
            }

            if (branchUntilNoNonCommons(vertices, index + 1, nonCommons, cost + costOfCaseB)) return true;

            for (int i = editsOfCasesB.length - 1; i >= 0; i--) {
                Edge e = edgesToNodes[i];
                if (blocksOfCasesB[i]) e.unblock();
                if (editsOfCasesB[i]) {
                    g.editEdge(e);
                }
            }
        }
        //end
        return false;
    }

    static Vertex[] getNonCommonArray(Graph g, Set<Vertex> verticesToMergeSet, List<Vertex> verticesToMergeList) {
        //set for constant contains request, list for better for-loop
        List<Vertex> vertices = new LinkedList<>();
        for (Vertex potentialNonCommon : g) {
            boolean isAdded = false;
            if (verticesToMergeSet.contains(potentialNonCommon)) continue;

            for (int i = 0; i < verticesToMergeList.size(); i++) {
                if (isAdded) break;//potentialNonCommon is a non common neighbor
                Vertex vToMerge = verticesToMergeList.get(i);
                for (int j = i + 1; j < verticesToMergeList.size(); j++) {
                    Vertex vToMerge2 = verticesToMergeList.get(j);
                    Edge e = g.getEdge(vToMerge, vToMerge2);
                    if (e.containsNonCommonNeighbor(potentialNonCommon)) {
                        vertices.add(potentialNonCommon);
                        isAdded = true;
                        break;
                    }
                }
            }
        }
        Vertex[] nonCommonArray = new Vertex[vertices.size()];
        vertices.toArray(nonCommonArray);
        return nonCommonArray;
    }

    static int minCostForResolvingNonCommons(Graph g, List<Vertex> vertices, Vertex[] nonCommonArray) {
        int minCostOfMerge = 0;
        for (Vertex nonCommon : nonCommonArray) {
            boolean hasBlockedEdge = false;
            boolean hasBlockedNonEdge = false;
            int costOfAdd = 0;
            int costOfRemove = 0;
            //cut branch off if there is a blocked edge and a blocked non-edge
            for (Vertex vertex : vertices) {
                Edge e = g.getEdge(vertex, nonCommon);
                if (e.isBlocked) {
                    if (e.isEdge) hasBlockedEdge = true;
                    else hasBlockedNonEdge = true;
                } else {
                    if (e.isEdge) costOfRemove += e.getEditCost();
                    else costOfAdd += e.getEditCost();
                }
                if (hasBlockedEdge && hasBlockedNonEdge) {
                    abortFlag = true;
                    break;
                }
            }
            if (abortFlag) break;
            //compute minimal cost
            if (hasBlockedEdge) {
                minCostOfMerge += costOfAdd;
            } else {
                if (hasBlockedNonEdge) minCostOfMerge += costOfRemove;
                else minCostOfMerge += Math.min(costOfRemove, costOfAdd);
            }
        }
        return minCostOfMerge;
    }

    static Set<Vertex> getAdditionalVerticesToMerge(Vertex u, Vertex v, Label label, boolean labelAsBool) {
        Set<Vertex> verticesToMerge = new HashSet<>();
        addAllVerticesToMerge(u, v, u.initialNC, label, verticesToMerge, labelAsBool);
        if (u.initialNC != v.initialNC)
            addAllVerticesToMerge(u, v, v.initialNC, label, verticesToMerge, labelAsBool);
        return verticesToMerge;
    }

    private static int addBetweenEdgesToEditAndReturnCost(Graph g, boolean clique, List<Vertex> verticesToMerge, int k, List<Edge> editedEdges) {
        int costOfMerge = 0;
        for (int i = 0; i < verticesToMerge.size(); i++) {
            Vertex u = verticesToMerge.get(i);
            for (int j = i + 1; j < verticesToMerge.size(); j++) {
                Vertex v = verticesToMerge.get(j);
                Edge e = g.getEdge(u, v);
                if (e.isEdge != clique) {
                    if (e.isBlocked) return k + 1;
                    costOfMerge += e.getEditCost();
                    if (costOfMerge > k) return costOfMerge;
                    editedEdges.add(e);
                }
            }
        }
        return costOfMerge;
    }

    private static void addAllVerticesToMerge(Vertex u, Vertex v, NC nc, Label label, Set<Vertex> verticesToMerge, boolean labelAsBool) {
        if (abortFlag) return;
        for (Vertex vertex : nc.vertices) {
            //get the merged parent-vertex of 'vertex'
            Vertex parent = vertex.getParentVertex();
            if (parent == u || parent == v) continue;

            if (parent.getLabel() == label) {
                Edge e1 = g.getEdge(parent, u);
                Edge e2 = g.getEdge(parent, v);
                if ((e1.isBlocked && e1.isEdge != labelAsBool)
                        || (e2.isBlocked && e2.isEdge != labelAsBool)) {//abort if they have conflicting adjacencies
                    abortFlag = true;
                    return;
                } else {
                    verticesToMerge.add(parent);
                }
            }
        }
    }
}