package algo.datastructure;


import java.util.*;

public class Block {

    private final int id;
    public int index;
    public int size = 0;

    public Vertex[] verticesArray = new Vertex[10];
    public int[] ncDistribution;

    static private int idCounter = 0;

    public Block(int initialNCsSize, int id) {
        ncDistribution = new int[initialNCsSize];
        index = id;
        this.id = idCounter++;
    }

    public Vertex get(int index) {
        return verticesArray[index];
    }

    public void addVertex(Vertex vertex) {
        ncDistribution[vertex.getIndexOfNC()]++;
        addVertexToArray(vertex);
    }

    public void removeVertex(Vertex vertex) {
        ncDistribution[vertex.getIndexOfNC()]--;
        removeVertexFromArray(vertex);
    }

    private void addVertexToArray(Vertex v) {
        expandIfNecessary();
        verticesArray[size] = v;
        v.indexOfBlockArrayPosition = size;
        size++;

        v.currentBlock = this;
    }

    private void removeVertexFromArray(Vertex v) {
        size--;
        if (size != v.indexOfBlockArrayPosition) {
            verticesArray[v.indexOfBlockArrayPosition] = verticesArray[size];
            verticesArray[v.indexOfBlockArrayPosition].indexOfBlockArrayPosition = v.indexOfBlockArrayPosition;
        }
        verticesArray[size] = null;
        v.indexOfBlockArrayPosition = -1;

        v.currentBlock = null;
    }

    public int size() {
        return size;
    }


    public void addAllVerticesOfNC(NC nc) {
        for (Vertex vertex : nc.vertices) addVertexToArray(vertex);
        ncDistribution[nc.indexForNCCostArray] += nc.size;
    }

    public void expandIfNecessary() {
        if (size == verticesArray.length) {
            Vertex[] old = verticesArray;
            verticesArray = new Vertex[(int) (size * 1.5)];
            System.arraycopy(old, 0, verticesArray, 0, size);
        }
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        Block block = (Block) o;
        return id == block.id;
    }

    @Override
    public int hashCode() {
        return Objects.hash(id);
    }
}
