package algo.datastructure;

import java.util.*;

public class NC {

    public PartSetVertices vertices;
    final int id;
    public int size = 0;
    public int indexForNCArray;
    public int indexForNCCostArray;

    public Label initialLabel = null;

    public NC(Graph g, int ncID) {
        id = ncID;
        vertices = new PartSetVertices(g);
        indexForNCCostArray = ncID;
        indexForNCArray = ncID;
    }

    public void addVertex(Vertex vertex) {
        vertices.add(vertex);
        size += vertex.weight;
        vertex.currentNC = this;
    }

    public void removeVertex(Vertex vertex) {
        vertices.remove(vertex);
        size -= vertex.weight;
        vertex.currentNC = null;
    }

    public Vertex getRepresentative() {
        return vertices.get(0);
    }

    public static List<NC> initializeNCs(Graph g) {
        List<NC> ncs = new ArrayList<>();
        for (Vertex vertex : g) {
            NC nc = findFitNC(vertex, ncs, g);
            if (nc == null) {
                nc = new NC(g, ncs.size());
                ncs.add(nc);
            }
            nc.addVertex(vertex);
            vertex.currentNC = nc;
            vertex.initialNC = nc;
        }
        for (NC nc : ncs) setInitialInsideAdjacency(g, nc);
        return ncs;
    }

    static private void setInitialInsideAdjacency(Graph g, NC nc) {
        if (nc.size == 1) {
            nc.initialLabel = Label.UNDECIDED;
            return;
        }
        Vertex u = nc.get(0);
        Vertex v = nc.get(1);
        nc.initialLabel = g.areNeighbors(u, v) ? Label.ADJACENT : Label.SEPARATE;
    }

    Vertex get(int index) {
        return vertices.get(index);
    }

    static private NC findFitNC(Vertex vertex, List<NC> ncs, Graph graph) {
        for (NC nc : ncs) {
            Vertex ncVertex = nc.getRepresentative();
            if (ncVertex.degree() != vertex.degree()) continue;

            boolean sameNeighborhood = true;
            for (Vertex neighbor : vertex) {
                if (neighbor == ncVertex) continue;
                if (!graph.areNeighbors(ncVertex, neighbor)) {
                    sameNeighborhood = false;
                    break;
                }
            }
            if (sameNeighborhood) return nc;
        }
        return null;
    }
}
