package algo.heuristic;

import algo.datastructure.*;
import algo.util.Util;

import java.util.*;
import java.util.stream.Collectors;

public class SplitHeuristic extends Util {

    private static final boolean improveByLocalSearch = true;

    public static BlockPartitionResult solve(Graph graph, int t) {
        final List<NC> initialNCs = NC.initializeNCs(graph);

        BlockPartition blockPartition = new BlockPartition(initialNCs);
        blockPartition.initializeUseOfEdgesFromVertexToBlock(graph);

        while (blockPartition.size() > t) {
            Block block = getBestBlockToSplit(blockPartition);
            blockPartition.splitBlock(block);
        }

        int cost;
        if (improveByLocalSearch) cost = LocalSearch.improve(graph, blockPartition, initialNCs);
        else cost = blockPartition.computeCost();

        for (Vertex vertex : graph) {
            vertex.currentBlock = null;
            vertex.indexOfBlockArrayPosition = -1;
        }
        return new BlockPartitionResult(blockPartition, cost);
    }

    private static Block getBestBlockToSplit(BlockPartition blockPartition) {
        Map<Block, Integer> bagCost = new HashMap<>();
        for (Block block : blockPartition.getBlocks()) {
            int costOfBag = 0;
            for (int i = 0; i < block.size; i++)
                costOfBag += blockPartition.computeBestBlocksByMovingVertex(block.get(i)).minCost;
            bagCost.put(block, costOfBag);
        }

        final int minCostIncrease = bagCost.values().stream().min(Integer::compareTo).get();
        //select random block with least costIncrease
        Block bestBlock = getRandomElementOf(bagCost.entrySet().stream().filter(x -> x.getValue() == minCostIncrease).collect(Collectors.toList())).getKey();
        log(String.format("bestBag %s wird ausgewählt mit Kosten %s", bestBlock.index, bagCost.get(bestBlock)));
        return bestBlock;
    }
}