package algo.util;

import java.util.Arrays;
import java.util.Comparator;

public class QuickSelect {

    //with integer array:
    public static void selectSortTopKMax(int[] arr, int topK) {
        selectMax(arr, 0, arr.length - 1, topK - 1);
    }

    static void selectMax(int[] arr, int start, int end, int topK) {
        if (start == end) {
            return;
        }
        int pivotIndex = partitionMax(arr, start, end);
        if (topK == pivotIndex) {
            return;
        } else if (topK < pivotIndex) selectMax(arr, start, pivotIndex - 1, topK);
        else selectMax(arr, pivotIndex + 1, end, topK);
    }

    static int partitionMax(int[] arr, int low, int high) {
        int pivot = arr[high];
        int i = (low - 1);
        for (int j = low; j < high; j++) {
            if (arr[j] >= pivot) {
                i++;
                swap(arr, i, j);
            }
        }
        swap(arr, i + 1, high);
        return i + 1;
    }

    static void swap(int[] arr, int i, int j) {
        int elem = arr[i];
        arr[i] = arr[j];
        arr[j] = elem;
    }

    //with generic array:
    public static <T> void selectSortTopKMax(T[] arr, int topK, int endOfArray, Comparator<T> comparator) {
        selectMax(arr, 0, endOfArray - 1, topK - 1, comparator);
    }

    static <T> void selectMax(T[] arr, int start, int end, int topK, Comparator<T> comparator) {
        if (start == end) {
            return;
        }

        int pivotIndex = partitionMax(arr, start, end, comparator);

        if (topK == pivotIndex) {
            return;
        } else if (topK < pivotIndex) selectMax(arr, start, pivotIndex - 1, topK, comparator);
        else selectMax(arr, pivotIndex + 1, end, topK, comparator);

    }

    static <T> int partitionMax(T[] arr, int low, int high, Comparator<T> comparator) {
        T pivot = arr[high];
        int i = (low - 1);
        for (int j = low; j < high; j++) {
            if (comparator.compare(arr[j], pivot) >= 0) {
                i++;
                swap(arr, i, j);
            }
        }
        swap(arr, i + 1, high);
        return i + 1;
    }

    static <T> void swap(T[] arr, int i, int j) {
        T elem = arr[i];
        arr[i] = arr[j];
        arr[j] = elem;
    }
}
